//professionals, close your eyes or be prepared with vomit pouch in hands!!  


#include <EEPROM.h>                //needed for store settings and antenna names

#include <LiquidCrystal.h>         //needed for interfacing with 16x2 chars LCD

#include <SoftwareSerial.h>        //used for CI-V reading

char inbyte;                       //several ancillary variables
char settext[10];
int eeadr = 0;
int buffget[11];
int rxbyte = 0;
int i = 0, j;
int band, pom, pomaux;
int relay;
int ant, lastant;
unsigned long watchdog = 0;
bool automatic = false;           //flag for automatic mode
bool wrong = false;               //flag for false reading of CI-V frequency commands
bool ssetup = false;              //flags for setup state-machine
bool namesetup = false;
bool switchsetup = false;
bool bylznak = false;
int unsigned long MHZ = 0;        //variables for translating CI-V format to real frequency reading
int unsigned long KHZ = 0;
int unsigned long HZ = 0;
int unsigned long QRG = 0;
// rotary encoder definitions follow
static int pinA = 2; // Our first hardware interrupt pin is digital pin 2
static int pinB = 3; // Our second hardware interrupt pin is digital pin 3
volatile byte aFlag = 0; // let's us know when we're expecting a rising edge on pinA to signal that the encoder has arrived at a detent
volatile byte bFlag = 0; // let's us know when we're expecting a rising edge on pinB to signal that the encoder has arrived at a detent (opposite direction to when aFlag is set)
volatile byte encoderPos = 0; //this variable stores our current value of encoder position. Change to int or uin16_t instead of byte if you want to record a larger range than 0-255
volatile byte oldEncPos = 0; //stores the last encoder position value so we can compare to the current reading and see if it has changed (so we know when to print to the serial monitor)
volatile byte reading = 0; //somewhere to store the direct values we read from our interrupt pins before checking to see if we have moved a whole detent

const byte antpin_1 = 4;   //which output pin switch which antenna relay definition
const byte antpin_2 = 5;
const byte antpin_3 = 6;
const byte antpin_4 = 7;
const byte antpin_5 = 10;
const byte antpin_6 = 11;
const byte auxpin = 12;     //auxiliary relay for connecting of AH-3 signal wires

char* bandString[]={" ", "160m", " 80m", " 60m"," 40m", " 30m"," 20m"," 17m", " 15m"," 12m", " 10m","  6m"};    //strings used as name of each band in setup



// Software serial routine setup
const byte rxPin = 8;         //this pin has to be connected to CI-V live pin through an cca 2k2 resistor which prevents affection of CI-V during uninitialized or unpowered state
const byte txPin = 9;         //not used in this application, we read only..
SoftwareSerial mySerial (rxPin, txPin);

// set up LCD
//LiquidCrystal lcd(rs, en, d4, d5, d6, d7);
LiquidCrystal lcd(A5, A4, A3, A2, A1, A0);       //pin assignment to LCD panel, 4-wire connection

void switchoff()              //routine for deactivating of all relay outputs
{
digitalWrite(antpin_1,LOW);
digitalWrite(antpin_2,LOW);
digitalWrite(antpin_3,LOW);
digitalWrite(antpin_4,LOW);
digitalWrite(antpin_5,LOW);
digitalWrite(antpin_6,LOW);
digitalWrite(auxpin,LOW);  
}

void vypisanteny()          //write list of antennas to serial port, used during setup
{
  Serial.println("Actual antennas setup:");
    Serial.println();
    Serial.println("Nr  Aux  Name");
    Serial.println("_____________");
    for ( i=1;i<7;i++)       //we have 6 relays (may be increased/decreased depending on actual switch setup)
      {
        eeadr = 16+(15*i);    //name of each antenna occupies 16 bytes starting from EEPROM address 16. First 16 bytes are used for band to antenna assignment. May be optimalized, but why....
        Serial.print(i);
        Serial.print(":  ");
        Serial.print(char(EEPROM.read(eeadr)));
        Serial.print("   ");
        for ( j=1;j<15;j++)
          {
            Serial.print(char(EEPROM.read(eeadr+j)));
           
          }
        Serial.println();
      }
    Serial.println();
}

void vypissetup()       //write list of actual band to antenna assignment
{
  Serial.println("Actual settings:");
    Serial.println();
    Serial.println("Band     Ant ");
    Serial.println("-------------");
     for ( i=1;i<12;i++)
      {
        pom = EEPROM.read(i);
         
        pom = pom & 0b00001111;
        Serial.print(i,HEX); Serial.print(": "); Serial.print(bandString[i]); Serial.print("   "); Serial.println(pom); ;
      }
    Serial.println();
}

void setup()    //setup routine, runs once after reset

{
  
  // for rotary encoder
  pinMode(pinA, INPUT_PULLUP); // set pinA as an input, pulled HIGH to the logic voltage (5V or 3.3V for most cases)
  pinMode(pinB, INPUT_PULLUP); // set pinB as an input, pulled HIGH to the logic voltage (5V or 3.3V for most cases)
  attachInterrupt(0,PinA,RISING); // set an interrupt on PinA, looking for a rising edge signal and executing the "PinA" Interrupt Service Routine (below)
  attachInterrupt(1,PinB,RISING); // set an interrupt on PinB, looking for a rising edge signal and executing the "PinB" Interrupt Service Routine (below)
 
 // for softserial
  pinMode(rxPin, INPUT);
  pinMode(txPin, OUTPUT);
  mySerial.begin(19200);      // modify to your ICOM CI-V speed. I recommend to change setup in TRX/CAT/LOG software to this speed, do not use higher due to software emulation of serial interface
  Serial.begin(38400);        // serial speed used for USB-serial interface. Used for setup.

  // set up the LCD's number of columns and rows:
  lcd.begin(16, 2);

  pinMode(antpin_1, OUTPUT);   // set all GPIO pins used for relays as outputs
  pinMode(antpin_2, OUTPUT);
  pinMode(antpin_3, OUTPUT);
  pinMode(antpin_4, OUTPUT);
  pinMode(antpin_5, OUTPUT);
  pinMode(antpin_6, OUTPUT);
  pinMode(auxpin, OUTPUT);

  digitalWrite(antpin_1,LOW);   // swicth everything off (may be optimalized, we have subroutine already defined so we can call it instead of this part of code)
  digitalWrite(antpin_2,LOW);
  digitalWrite(antpin_3,LOW);
  digitalWrite(antpin_4,LOW);
  digitalWrite(antpin_5,LOW);
  digitalWrite(antpin_6,LOW);
  digitalWrite(auxpin,LOW);
}

void PinA()          //incremental encoder interrupt routines
{
  cli(); //stop interrupts happening before we read pin values
  reading = PIND & 0xC; // read all eight pin values then strip away all but pinA and pinB's values
  if(reading == B00001100 && aFlag) { //check that we have both pins at detent (HIGH) and that we are expecting detent on this pin's rising edge
    encoderPos ++; //decrement the encoder's position count
    automatic = false;
    
    if (encoderPos == 7) encoderPos = 0;
    if (encoderPos == 255) encoderPos = 6;
    
    bFlag = 0; //reset flags for the next turn
    aFlag = 0; //reset flags for the next turn
  }
  else if (reading == B00000100) bFlag = 1; //signal that we're expecting pinB to signal the transition to detent from free rotation
  sei(); //restart interrupts
}

void PinB()
{
  cli(); //stop interrupts happening before we read pin values
  reading = PIND & 0xC; //read all eight pin values then strip away all but pinA and pinB's values
  if (reading == B00001100 && bFlag) { //check that we have both pins at detent (HIGH) and that we are expecting detent on this pin's rising edge
    encoderPos --; //increment the encoder's position count
    automatic = false;
    
    if (encoderPos == 7) encoderPos = 0;
    if (encoderPos == 255) encoderPos = 6;
    
    bFlag = 0; //reset flags for the next turn
    aFlag = 0; //reset flags for the next turn
  }
  else if (reading == B00001000) aFlag = 1; //signal that we're expecting pinA to signal the transition to detent from free rotation
  sei(); //restart interrupts
}


void loop()                 //main loop

{

if (Serial.available() > 0)       //have an char at USB-serial interface?
  {
    inbyte = Serial.read();
    if (inbyte == 0x0D) ssetup = true;            //if (ENTER) was pressed, do setup
  }

if (ssetup)
  {
    
    lcd.clear();
    lcd.write("Setup entered");
    Serial.println();
    Serial.println();
    Serial.println("Setup entered!");
    Serial.println();
    vypisanteny();
    vypissetup();
  } 
   
while (ssetup)    
  {  
    Serial.print("Press (N)ame / (S)et / e(X)it :");
    
    do
      {
        if (Serial.available() > 0) 
          {
            inbyte = Serial.read();
            Serial.println(inbyte);
          }
        
      }
    
    while (inbyte != 'n' && inbyte != 'N' && inbyte != 's' && inbyte != 'S' && inbyte != 'x' && inbyte != 'X');       //do not react to another chars than defined

    
    if (inbyte == 'x' || inbyte == 'X')
      {
        ssetup = false;    //when x or X pressed, leave setup
        Serial.println();
        Serial.println();
        Serial.println("********  Leaving setup.  73! CUAGN  ************");
      }

    
    if (inbyte == 'n' || inbyte == 'N')                    //antenna name setup
      {
        namesetup = true;
        Serial.print("Enter number of antenna or e(X)it :");
        while (namesetup)
         {
           do
            {
              if (Serial.available() > 0) 
                {
                  inbyte = Serial.read();
                  Serial.println(inbyte);
                }
        
            }  while (inbyte != '1' && inbyte != '2' && inbyte != '3' && inbyte != '4' && inbyte != '5' && inbyte != '6' && inbyte != 'x' && inbyte != 'X');   //do not react to another chars than defined

           
           if (inbyte == 'x' || inbyte == 'X') 
              {
                inbyte = 0;
                namesetup = false;
                                           //leave this level of setup menu
              }
           Serial.println();
           
           if (inbyte == '1' || inbyte == '2' || inbyte == '3' || inbyte == '4' || inbyte == '5' || inbyte == '6')
            { 
           
           pom = inbyte-48;
           eeadr = 16 + (15*pom);
           Serial.print(pom); Serial.print(". antenna name (first char 0 = Aux off 1 = Aux on): ");         //first char defines state of AUX relay
           
           for ( j=0;j<15;j++)                                 //fill all chars, add (SPACE) until line is finished
            {
              bylznak = false;
              do
              {
                if (Serial.available() > 0) 
                 {
                    inbyte = Serial.read();                     //
                    bylznak = true;
                  }
        
              }  while (bylznak == false);

              EEPROM.write((eeadr + j),inbyte);                  //write string to EEPROM
              Serial.print(inbyte);
            }
           Serial.println("   Done!");
           vypisanteny();
           Serial.print("Enter number of antenna or e(X)it :");      //another one or exit?
            }

           
         }
      }



      
     if (inbyte == 's' || inbyte == 'S')                    //band to relay setup
      {
        switchsetup = true;
        vypissetup();
        Serial.print("Enter band (1..B) or e(X)it :");       //for simplicity we choose band by pressing one char. We have 11 bands, so A = 10, B = 11
        while (switchsetup)
         {
           do
            {
              if (Serial.available() > 0) 
                {
                  inbyte = Serial.read();
                  
                }
        
            }  while (inbyte != '1' && inbyte != '2' && inbyte != '3' && inbyte != '4' && inbyte != '5' && inbyte != '6' && inbyte != '7' && inbyte != '8' && inbyte != '9' && inbyte != 'A' && inbyte != 'B'  && inbyte != 'a' && inbyte != 'b' && inbyte != 'x' && inbyte != 'X');

           
           if (inbyte == 'x' || inbyte == 'X') 
              {
                inbyte = 0;
                switchsetup = false;
                                           //leave this level of setup
              }
           
           
           if (inbyte == '1' || inbyte == '2' || inbyte == '3' || inbyte == '4' || inbyte == '5' || inbyte == '6' || inbyte == '7' || inbyte == '8' || inbyte == '9' || inbyte == 'A' || inbyte == 'B' || inbyte == 'a' || inbyte == 'b')
            { 
              Serial.println(inbyte);
              if (inbyte == 'A' || inbyte == 'a') pom = 10;
              else 
                { 
                  if (inbyte == 'B' || inbyte == 'b') pom = 11; 
                  else pom = inbyte-48;
                }
           
              Serial.print(bandString[pom]); Serial.print(" which antenna?: ");
           
          
              bylznak = false;
              do
              {
                if (Serial.available() > 0) 
                 {
                    inbyte = Serial.read();
                    bylznak = true;
                  }
        
              }  while (bylznak == false);

              EEPROM.write((pom),inbyte);
              
              Serial.print(inbyte);
              inbyte = 0;
              Serial.println("   Done!");
              vypissetup();
              Serial.print("Enter band (1..B) or e(X)it :");
            }
    
         }
      }  
  }

if (encoderPos == 0) automatic = true;             //when encoder position = 0, do automatic recognition of band

 if ((automatic)&&(mySerial.available() > 0)&&(!ssetup))
 { 
  while (rxbyte != 0xFE)    //wait for CI-V prefix
  { 
    if (!automatic) break;
    if (mySerial.available() > 0) 
    {
      rxbyte = mySerial.read();
      watchdog = millis();
    }
    if (!automatic) break;
    if ((millis() - watchdog) > 1000) {  break;}  //timeout, when no CI-V read, allow manual entry
  }
 delay(10);
  for ( i=0;i<11;i++)
  {               
    while (mySerial.available() == 0) 
    {
      if (!automatic) break;
      if ((millis() - watchdog) > 1000) break;
    }
    buffget[i] = mySerial.read();
    
    watchdog = millis();
    if (buffget[i] == 0xFD) break;      //end of CI-V message
  } 
  
  if (((buffget[4] == 0) or (buffget[4] == 3) or (buffget[4] == 5)) and (buffget[5] != 0xFD))   //is this command containing frequency information??
 
  {
    
    MHZ = (buffget[8]);
    MHZ = MHZ - (((MHZ/16) * 6));              // Transform bytes ICOM CAT 
    if (MHZ >= 100)  wrong = true;              // wrong byte

    KHZ = buffget[7];
    KHZ = KHZ - (((KHZ/16) * 6));              // Transform bytes ICOM CAT 
    if (KHZ >= 100)  wrong = true;              // wrong byte

    HZ = buffget[6]; 
    HZ = HZ - (((HZ/16) * 6));                 // Transform bytes ICOM CAT 
    if (HZ >= 100)  wrong = true;               // wrong byte

    if (!wrong) QRG = ((MHZ * 10000) + (KHZ * 100) + (HZ * 1))/10;
    else wrong = false; // QRG variable stores frequency in MMkkkH  format
  
  if ((QRG > 1500) && (QRG <= 2100)) band = 1;         //rough limits between which valid band is recognized
  if ((QRG > 3400) && (QRG <= 4000)) band = 2;
  if ((QRG > 5000) && (QRG <= 6000)) band = 3;
  if ((QRG > 6900) && (QRG <= 7500)) band = 4;
  if ((QRG > 9100) && (QRG <= 10500)) band = 5;
  if ((QRG > 13900) && (QRG <= 15000)) band = 6;
  if ((QRG > 18000) && (QRG <= 19000)) band = 7;
  if ((QRG > 20000) && (QRG <= 23000)) band = 8;
  if ((QRG > 24000) && (QRG <= 26000)) band = 9;
  if ((QRG > 27000) && (QRG <= 31000)) band = 10;
  if ((QRG > 49000) && (QRG <= 55000)) band = 11;
    
  
  ant = EEPROM.read(band);                        //read EEPROM, which switch is defined for rocognized band
  
  
  lcd.setCursor(0, 0);                           //write QRG info to LCD, first line
  lcd.print("QRG: ");
  if (QRG < 10000) lcd.print(" ");
  lcd.print(QRG/1000);
  lcd.print(".");
  if (QRG%1000 < 100) lcd.print("0");
  if (QRG%1000 < 10) lcd.print("0");
  lcd.print(QRG%1000);
  lcd.print(" MHz ");
  
  }


  
 }
  
  if (!automatic)                               //write "Manual mode" when is not automatic mode
  {
    lcd.setCursor(0, 0);
    lcd.print("Manual mode     ");
    ant = encoderPos;
  }
  
  
                                              //was band changed?
  if (ant != lastant)
    {
      lastant == ant;
      switchoff();                           //firstly switch all off
      
      ant = ant & 0b00001111;
      if (ant == 1) digitalWrite(antpin_1,HIGH);          //switch one output
      if (ant == 2) digitalWrite(antpin_2,HIGH);
      if (ant == 3) digitalWrite(antpin_3,HIGH);
      if (ant == 4) digitalWrite(antpin_4,HIGH);
      if (ant == 5) digitalWrite(antpin_5,HIGH);
      if (ant == 6) digitalWrite(antpin_6,HIGH);

      lcd.setCursor(0, 1);                               //set second row on LCD
      
      lcd.print(ant);                                    //write number of antenna
      lcd.print(" ");
      eeadr = 16+(ant*15);
      if (EEPROM.read(eeadr) == '1') digitalWrite(auxpin,HIGH);       //look if AUX relay have to be switched on or not
      for ( i=1;i<15;i++)
        {
          lcd.print(char(EEPROM.read(eeadr+i)));                    //write name of antenna
        }

      
    }
  {
    
  }

}

// Thats all, folk!

// Yes, I know that code is ugly, poorly commented, highly inefficient.. But note, I am not professional programmer, only hobbyist ;)

// This code is writen using my own effort and ideas or chunks of code of another authors which are freely available across arduino community.
// You can use this code as is, modify it, share it to friends, but please do not use it in any kind of commercial projects without my permission.
// No warranty for correct function. Do not use it in life critical applications :o)
// (c) OK2WY 2017